// HanAutomata.cpp: implementation of the _CHanAutomata class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "IMEsebol.h"
#include "HanAutomata.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHanAutomata::CHanAutomata()
 : cho(m_jamo[0]), jung(m_jamo[1]), jong(m_jamo[2])
{
	clear();
}

CHanAutomata::~CHanAutomata()
{

}

void CHanAutomata::clear()
{
	memset(m_szBuf, 0, sizeof (m_szBuf));
	m_curr = 0;
	m_bul = BLS_NULL;

	cho = 0xFF;
	jung = 0xFF;
	jong = 0;
}

int CHanAutomata::setString(LPCTSTR pszStr)
{
	clear();
	for (LPCTSTR p = pszStr; *p; p++) {
		addChar((char)*p);
	}

	return 0;
}

int CHanAutomata::getUnicode(WCHAR *pBuf)
{
	lstrcpyW(pBuf, m_szBuf);
	int len = lstrlenW(pBuf);

	WCHAR wz[2] = L"x";
	wz[0] = makeUnicode();

	if (wz[0]) {
		lstrcatW(pBuf, wz);
		len++;
	}

	return len;
}

int CHanAutomata::addChar(char ch)
{
	WORD code = getCode(ch);
	BYTE bul = HIBYTE(code);
	BYTE cod = LOBYTE(code);

	if (code == 0xFFFF || bul > 2) { // ڵǥ ų ѱ ƴϸ
		if (m_bul != BLS_NULL) {
			complete();
		}

		m_szBuf[m_curr++] = (WORD)(BYTE)(code == 0xFFFF ? ch : cod);
		return BLS_NULL;
	}

	if (m_bul == BLS_NULL || bul == m_bul + 1) {
		//  ڰ ų 
		// ʼ ߼, ߼  
		m_jamo[bul] = cod;
		return m_bul = bul;
	}

	if (bul == m_bul) { // // ޾Ƽ  
		int iStart = bul == 0 ? 0 : bul == 1 ? 5 : 12;
		for (int i = iStart; s_cnvtbl[i][0] == bul; i++) {
			if (m_jamo[bul] == s_cnvtbl[i][1] && cod == s_cnvtbl[i][2]) {
				m_jamo[bul] = s_cnvtbl[i][3];
				return bul;
			}
		}
	}

	complete();

	m_jamo[bul] = cod;
	return m_bul = bul;
}

WORD CHanAutomata::getCode(char ch)
{
	for (const _KeyCode *p = s_3bulKeys; p->ch; p++) {
		if (p->ch == ch)
			return MAKEWORD(p->cod, p->bul);
	}
	return 0xFFFF;
}

void CHanAutomata::complete()
{
	WORD code = makeUnicode();;

	if (code != 0) {
		m_szBuf[m_curr] = code;
		m_curr++;
	}

	cho = 0xFF, jung = 0xFF, jong = 0;
	m_bul = BLS_NULL;
}

WORD CHanAutomata::makeUnicode()
{
	if (cho != 0xFF && jung != 0xFF) {
		return cho * (21 * 28) + jung * 28 + jong + 0xAC00;
	}

	if (cho != 0xFF && jung == 0xFF && jong == 0) {
		return 0x1100 + cho;
	}

	if (cho == 0xFF && jung != 0xFF && jong == 0) {
		return 0x1161 + jung;
	}

	if (cho == 0xFF && jung == 0xFF && jong != 0) {
		return 0x11A7 + jong;
	}

	if (cho != 0xFF || jung != 0xFF || jong != 0) {
		return 0x003F; // '?'
	}

	return 0;
}

const CHanAutomata::_KeyCode CHanAutomata::s_3bulKeys[] = {
//  0  1  2  3  4  5  6  7  8  9
// 10 11 12 13 14 15 16 17 18
	{ 'k',  0, 0, 0 },
	{ 'h',  2, 0, 0 },
	{ 'u',  3, 0, 0 },
	{ 'y',  5, 0, 0 },
	{ 'i',  6, 0, 0 },
	{ ';',  7, 0, 0 },
	{ 'n',  9, 0, 0 },
	{ 'j', 11, 0, 0 },
	{ 'l', 12, 0, 0 },
	{ 'o', 14, 0, 0 },
	{ '0', 15, 0, 0 },
	{ '\'',16, 0, 0 },
	{ 'p', 17, 0, 0 },
	{ 'm', 18, 0, 0 },

//  0  1  2  3  4  5  6  7  8  9
// 10 11 12 13 14 15 16 17 18 19 20
	{ 'f',  0, 1, 0 },
	{ 'r',  1, 1, 0 },
	{ '6',  2, 1, 0 },
	{ 'R',  3, 1, 0 },
	{ 't',  4, 1, 0 },
	{ 'c',  5, 1, 0 },
	{ 'e',  6, 1, 0 },
	{ '7',  7, 1, 0 },
	{ 'v',  8, 1, 0 },
	{ '/',  8, 1, 0 },
	{ '4', 12, 1, 0 },
	{ 'b', 13, 1, 0 },
	{ '9', 13, 1, 0 },
	{ '5', 17, 1, 0 },
	{ 'g', 18, 1, 0 },
	{ '8', 19, 1, 0 },
	{ 'd', 20, 1, 0 },

//  1  2  3  4  5  6  7  8  9 10
// 11 12 13 14 15 16 17 18 19 20
// 21 22 23 24 25 26 27
	{ 'x',  1, 2, 0 },
	{ 'F',  2, 2, 0 },
	{ 's',  4, 2, 0 },
	{ 'S',  6, 2, 0 },
	{ 'A',  7, 2, 0 },
	{ 'w',  8, 2, 0 },
	{ 'D',  9, 2, 0 },
	{ 'C', 10, 2, 0 },
	{ 'V', 15, 2, 0 },
	{ 'z', 16, 2, 0 },
	{ '3', 17, 2, 0 },
	{ 'X', 18, 2, 0 },
	{ 'q', 19, 2, 0 },
	{ '2', 20, 2, 0 },
	{ 'a', 21, 2, 0 },
	{ '!', 22, 2, 0 },
	{ 'Z', 23, 2, 0 },
	{ 'E', 24, 2, 0 },
	{ 'W', 25, 2, 0 },
	{ 'Q', 26, 2, 0 },
	{ '1', 27, 2, 0 },

// special characters
	{ 'N', '0', 3, 0 },
	{ 'M', '1', 3, 0 },
	{ '<', '2', 3, 0 },
	{ '>', '3', 3, 0 },
	{ 'J', '4', 3, 0 },
	{ 'K', '5', 3, 0 },
	{ 'L', '6', 3, 0 },
	{ 'U', '7', 3, 0 },
	{ 'I', '8', 3, 0 },
	{ 'O', '9', 3, 0 },
	{ 'Y', '<', 3, 0 },
	{ 'P', '>', 3, 0 },
	{ 'H', '\'',3, 0 },
	{ 'G', '/', 3, 0 },
	{ 'T', ';', 3, 0 },
	{ 'B', '!', 3, 0 },
// end
	{   0,  0, 0, 0 },
};

const CHanAutomata::_KeyCode CHanAutomata::s_2bulKeys[] = {
// end
	{   0,  0, 0, 0 },
};


const BYTE CHanAutomata::s_cnvtbl[][4] = {
//  0  1  2  3  4  5  6  7  8  9
// 10 11 12 13 14 15 16 17 18
	{ 0,  0,  0,  1},
	{ 0,  3,  3,  4},
	{ 0,  7,  7,  8},
	{ 0,  9,  9, 10},
	{ 0, 12, 12, 13},
//  0  1  2  3  4  5  6  7  8  9
// 10 11 12 13 14 15 16 17 18 19 20
	{ 1,  8,  0,  9},
	{ 1,  8,  1, 10},
	{ 1,  8, 20, 11},
	{ 1, 13,  4, 14},
	{ 1, 13,  5, 15},
	{ 1, 13, 20, 16},
	{ 1, 18, 20, 19},
//  1  2  3  4  5  6  7  8  9 10
// 11 12 13 14 15 16 17 18 19 20
// 21 22 23 24 25 26 27
	{ 2,  1,  1,  2},
	{ 2,  1, 19,  3},
	{ 2,  4, 22,  5},
	{ 2,  4, 27,  6},
	{ 2,  8,  1,  9},
	{ 2,  8, 16, 10},
	{ 2,  8, 17, 11},
	{ 2,  8, 19, 12},
	{ 2,  8, 25, 13},
	{ 2,  8, 26, 14},
	{ 2,  8, 27, 15},
	{ 2, 17, 19, 18},
	{ 2, 19, 19, 20},
// end
	{ 0,  0,  0,  0}
};

